#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed Dec. 2007 by Fredo6

# This software is provided as an example of using the Ruby interface to SketchUp.

# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name		:   Shear.rb
# Type		:   Tool 
# Description	:   shear elements ofthe selection by a reference direction and a rotation vector.
# Menu Item	:   1 menu items "Shear Transformation" in Tools menu of Sketchup
# Context Menu	:   none
# Usage		:   See Tutorial 
# Date		:   04 Dec 2007
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
require 'sketchup.rb'
require 'LibTraductor.rb'		# for language translation
require 'LibProtractorTool.rb'	# for Protractor tool to input the origin, and direction vectors

module NULTransformation

STR_Title = ["Shear Transformation",
             "|FR| Cisaillement plan"]
STR_Tooltip = ["Deform a rectangle into a parallelogram",
               "|FR| D\forme un rectangle en parall\logramme"]
STR_MsgNoSelection = ["You must select at least one element",
                      "|FR| Aucun element selectionn\"]
STR_MsgAngleZero = ["No Shear to perform - Angle is ZERO",
                    "|FR| Angle Nul - Pas de cisaillement possible"]
STR_MsgAngleHigh = ["Angle is too close to 90 degree: %1",
                    "|FR| Angle trop proche de 90 degr\: %1"]
STR_MsgPB_Completion = ["Elts:",
                        "|FR| Elts:"]

NULT_TOOLBAR = "NUL Transformation"

#********************************************************************
# BASE CLASS for processing transformation of entities.
# The class uses a method <Transform_point(pt)> that must be implemneted by the
#  actual class defining the method
#********************************************************************

class PointBasedTransformation

def init_context
	@model = Sketchup.active_model
    @entities = @model.active_entities
    @selection = @model.selection
	@nbelts = 0
	@uv0 = Geom::Point3d.new 0, 0, 0
	@uv1 = Geom::Point3d.new 1, 0, 0
	@uv2 = Geom::Point3d.new 1, 1, 0
	@uv3 = Geom::Point3d.new 0, 1, 0
	@tw = Sketchup.create_texture_writer
	@tr_identity = Geom::Transformation.new
end

def process_entities(selection)
	init_context
	count_entities selection
	@pbar = ProgressionBar.new @nbelts, STR_MsgPB_Completion
	@model.start_operation Traductor[STR_Title]
		loop_entities @entities.add_group, selection, @tr_identity	
		@entities.erase_entities selection   #Removing all remaining entities
	@model.commit_operation	
	@model.select_tool nil
	return true			# to tell the Protractor this is OK
end

#Count all entities in the selection
def count_entities(selection)
	selection.each do |entity| 
		case entity.typename
		when 'Edge'
			@nbelts += 1 unless entity.faces
		when 'Face'
			@nbelts += entity.edges.length
		when 'ConstructionPoint', 'ConstructionLine', 'Text', 'Image'
			@nbelts += 1
		when 'Group'
			count_entities entity.entities
		when 'ComponentInstance'
			count_entities entity.definition.entities
		else
			selection.remove entity
		end	
	end	
end

#Recursive function to loop thorugh all the entities contained in <selection> and create the new geometry into <group>
# <t> is accumulated transformation from getting into groups and components, if any
def loop_entities(group, selection, t)
	hashedges = {}
	hashcurves = {}
	
	#We recurse on Groups and components
	selection.each do |entity| 
		case entity.typename
		when 'Group', 'ComponentInstance'
			e = (entity.typename == 'Group') ? entity : entity.definition
			newgrp = group.entities.add_group
			newgrp.name = e.name
			newgrp.description = e.description
			newgrp.locked = entity.locked?
			transfer_drawing_element e, newgrp
			transfer_attributes entity, newgrp
			loop_entities newgrp, e.entities, t * entity.transformation
		when 'Face'
			compute_face entity, group, t
			entity.edges.each {|e| hashedges[e.to_s] = e unless e.curve}
		when 'Edge'
			curve = entity.curve
			if curve
				hashcurves[curve.to_s] = curve
			else
				hashedges[entity.to_s] = entity
			end	
		when 'ConstructionLine'
			compute_cline entity, group, t
		when 'ConstructionPoint'
			compute_cpoint entity, group, t
		when 'Text'
			compute_text entity, group, t
		when 'Image'
			compute_image entity, group, t
		end	
	end	
	
	#We manage the edges and then curves
	hashedges.each {|key, edge| compute_edge edge, group, t}	
	hashcurves.each {|key, curve| compute_curve curve, group, t}
end

def edge_for_draw(selection, t, pts)
	selection.each do |entity| 
		case entity.typename
		when 'Group', 'ComponentInstance'
			e = (entity.typename == 'Group') ? entity : entity.definition
			edge_for_draw e.entities, t * entity.transformation, pts
		when 'Edge'
			entity.vertices.each {|v| pts.push transform_point(t * v.position)}
		end	
	end	
end

def compute_face(face, grp, t)
	return unless face && face.loops
	face.loops.each do |l|
		pts = compute_vertices l.vertices, t
		pts.push pts[0].clone
		newface = grp.entities.add_face pts
		next unless newface
		unless l.outer?
			grp.entities.add_edges pts
			grp.entities.erase_entities newface
		else
			newface.reverse! if (face.normal.transform(t) % newface.normal < 0)
			transfer_drawing_element face, newface
			newface.back_material = face.back_material
			#l.edges.each {|e| e.hidden = true}
			transfer_texture face, newface, true
			transfer_texture face, newface, false
		end
		@pbar.countage l.edges.length
	end			
end

def transfer_texture(oldface, newface, front)
	m = (front) ? oldface.material : oldface.back_material
	return unless m && m.texture
	uvh = oldface.get_UVHelper front, !front, @tw
	nv = oldface.outer_loop.vertices.length - 1
	ptuv = []
	oldptuv = []
	j = 0
	for i in 0..nv
		oldpt = oldface.outer_loop.vertices[i].position
		next if (j >= 2) && (oldpt.on_line? [oldptuv[j-2], oldptuv[j-1]])
		oldptuv.push oldpt
		j += 1
		ptuv.push newface.outer_loop.vertices[i].position
		if (front)
			ptuv.push uvh.get_front_UVQ(oldpt)
		else
			ptuv.push uvh.get_back_UVQ(oldpt)
		end
		break if (j == 4)
	end
	begin
		newface.position_material m, ptuv, front
	rescue		#apparently some bugs in Sketchup with projected textures!
	end
end

def compute_edge(edge, grp, t)
	pts = compute_vertices edge.vertices, t
	newedges = grp.entities.add_edges pts	
	transfer_edge(edge, newedges[0])
	@pbar.countage unless edge.faces
end

def transfer_edge(edge, newedge)
	newedge.smooth = edge.smooth?
	newedge.soft = edge.soft?
	transfer_drawing_element edge, newedge
end

def transfer_drawing_element (old_entity, new_entity)
	new_entity.layer = old_entity.layer
	new_entity.material = old_entity.material
	new_entity.visible = old_entity.visible?
	new_entity.receives_shadows = old_entity.receives_shadows?
	new_entity.casts_shadows = old_entity.casts_shadows?
end

def compute_image(image, grp, t)
	origin = image.origin
	w = image.width
	h = image.height
	angle = Math::atan2 h, w
	vec = origin.vector_to image.bounds.center
	vecx = Geom::Transformation.rotation(origin, image.normal, -angle) * vec
	vecy = Geom::Transformation.rotation(origin, image.normal, Math::PI * 0.5 - angle) * vec
	ptx = origin.offset vecx, w
	pty = origin.offset vecy, h
	ptxy = pty.offset vecx, w
	pt0 = transform_point(t * origin)
	pt1 = transform_point(t * ptx)
	pt2 = transform_point(t * ptxy)
	pt3 = transform_point(t * pty)
	pt4 = pt0.clone
	face = grp.entities.add_face [pt0, pt1, pt2, pt3, pt4]
	face.reverse! unless image.normal.samedirection? face.normal
	transfer_drawing_element image, face
	
	#Assigning and deforming the image texture
	m = Sketchup.active_model.materials.add ('nult_' + File.basename(image.path))
	m.texture = image.path
	m.texture.size = w
	face.position_material m, [pt0, @uv0, pt1, @uv1, pt2, @uv2, pt3, @uv3], true 
	face.position_material m, [pt0, @uv0, pt1, @uv1, pt2, @uv2, pt3, @uv3], false 
	@pbar.countage
end

def compute_cline(cline, grp, t)
	ptpos = transform_point(t * cline.position)
	ptpos2 = transform_point(t * cline.position.offset(cline.direction, 10))
	newcline = grp.entities.add_cline ptpos, ptpos2, cline.stipple
	newcline.start = (cline.start) ? transform_point(t * cline.start) : nil
	newcline.end = (cline.end) ? transform_point(t * cline.end) : nil
	transfer_drawing_element cline, newcline	
	transfer_attributes cline, newcline
	@pbar.countage
end

def compute_cpoint(cpoint, grp, t)
	pt = transform_point(t * cpoint.position)
	newcpoint = grp.entities.add_cpoint pt
	transfer_drawing_element cpoint, newcpoint	
	transfer_attributes cpoint, newcpoint	
	@pbar.countage
end

def compute_text (ctext, grp, t)
	pt = transform_point(t * ctext.point)
	pt2 = transform_point(t * ctext.point.offset(ctext.vector, 10))
	newctext = grp.entities.add_text ctext.text, pt, pt.vector_to(pt2)
	newctext.arrow_type = ctext.arrow_type
	newctext.display_leader = ctext.display_leader?
	newctext.leader_type = ctext.leader_type
	newctext.line_weight = ctext.line_weight
	transfer_drawing_element ctext, newctext	
	transfer_attributes ctext, newctext	
end

def compute_curve(curve, grp, t)
	pts = compute_vertices curve.vertices, t
	g = grp.entities.add_group		#only way to mkae sure we create the curve is to do it in a group
	g.entities.add_curve pts
	g.explode.each do |edge|		#retrieve the curve after explosion of the temporary group
		if (edge.typename == 'Edge')
			edge.curve.edges.each {|e| transfer_edge curve.first_edge, e}
			newcurve = edge.curve
			transfer_attributes curve, newcurve
			break
		end
	end	
	#transfer_attributes curve, newcurve
	@pbar.countage curve.edges.length unless curve.edges[0].faces
end

def transfer_attributes oldentity, newentity
	dd = oldentity.attribute_dictionaries 
	return unless dd
	dd.each do |d|
		d.each do |key, value|
			newentity.set_attribute d.name, key, value
			if d.name == "skp" && key == "crvpts"
				pts = []
				value.each { |pt| pts.push transform_point(pt) }
				newentity.set_attribute d.name, key, pts
			end
		end
	end
end

#This is the routine that computes the new coordinates of vertices  
def compute_vertices (vertices, t)
    pts = []
	vertices.each {|v| pts.push transform_point(t * v.position)}
	pts
end

end #Class PointBasedTransformation

#--------------------------------------------------------------------------------------------------------------
# Class ProgressionBar: progress bar in the Sketchup Status text area
#--------------------------------------------------------------------------------------------------------------
class ProgressionBar

#Initialization of progress bar
def initialize(nbelts, label)
	@pb_nbelts = nbelts
	@pb_label = Traductor[label]
	@pb_progression = 0
	@pb_rangemax = 200
	@pb_range = 0
	@pb_time0 = Time.now	
end

#Increment the Progression Bar by <nb> steps
def countage(nb=1)
	@pb_progression += nb
	f = 100 * @pb_progression / @pb_nbelts
	percent = f.to_i
	if (percent != @pb_range)
		@pb_range = percent
		n = percent * @pb_rangemax / 100
		Sketchup::set_status_text "|" * n.to_i
	end	
	Sketchup.set_status_text @pb_label + " #{@pb_progression} / #{@pb_nbelts}", SB_VCB_LABEL
	Sketchup::set_status_text "#{@pb_range}%  -  #{sprintf "%4.2f", Time.now - @pb_time0} sec", SB_VCB_VALUE
end

end #class ProgressionBar

#********************************************************************
# SUB CLASS for processing Plane Shear of entities.
# The class implemnets the method <Transform_point(pt)> to shear a point
#********************************************************************

class PlaneShear < PointBasedTransformation

#Call back Interface for execution from the Protractor tool
def protractor_execute (origin, vecref, vecdir)
	return false unless check_parameters(origin, vecref, vecdir, true)
	process_entities Sketchup.active_model.selection
	Sketchup.active_model.select_tool nil
	return true			# to tell the Protractor this is OK
end

#Call back Interface for interactive draw from the Protractor tool
def protractor_draw (view, origin, vecref, vecdir)
	pts = []
	return unless (vecdir && vecdir.valid? == true && check_parameters(origin, vecref, vecdir, true))
	edge_for_draw Sketchup.active_model.selection, Geom::Transformation.new, pts
	view.line_stipple = ""
	view.drawing_color = "purple"
	view.draw_lines pts
end

def check_parameters(origin, vecref, vecdir, silent)
	#Setting te geometry: origin, direction and rotation vectors
	@origin = origin
	@vecref = vecref
	@vecdir = vecdir
	
	#check that the angle is valid
	angle = @vecref.angle_between(@vecdir)
	sinus = Math.sin angle
	if (sinus == 0.0)
		UI.messagebox Traductor[STR_MsgAngleZero] unless silent
		return false
	elsif (sinus.abs > 0.95)
		UI.messagebox Traductor.s(STR_MsgAngleHigh, sprintf("%3.1f", angle * 180.0 / Math::PI)) unless silent
		return false
	end
	
	#computing the key parameter of the shear transformation
	@tangent = Math.tan angle
	@plane_ortho = [@origin, @vecref]
	pt0 = @origin.offset @vecdir, 10
	pt1 = pt0.project_to_plane @plane_ortho
	@vec_offset = Geom::Vector3d.new pt1.x - @origin.x, pt1.y - @origin.y, pt1.z - @origin.z

	return true
end

def process_shearing(origin, vecref, vecdir)
	#Setting te geometry: origin, direction and rotation vectors
	@origin = origin
	@vecref = vecref
	@vecdir = vecdir
	
	#check that the angle is valid
	angle = @vecref.angle_between(@vecdir)
	sinus = Math.sin angle
	if (sinus == 0.0)
		UI.messagebox Traductor[STR_MsgAngleZero]
		return false
	elsif (sinus.abs > 0.95)
		UI.messagebox Traductor.s(STR_MsgAngleHigh, sprintf("%3.1f", angle * 180.0 / Math::PI))
		return false
	end
	
	#computing the key parameter of the shear transformation
	@tangent = Math.tan angle
	@plane_ortho = [@origin, @vecref]
	pt0 = @origin.offset @vecdir, 10
	pt1 = pt0.project_to_plane @plane_ortho
	@vec_offset = Geom::Vector3d.new pt1.x - @origin.x, pt1.y - @origin.y, pt1.z - @origin.z
	
	#shearing the Selection
	process_entities Sketchup.active_model.selection
	Sketchup.active_model.select_tool nil
	return true			# to tell the Protractor this is OK
end

def oldprocess_shearing(origin, vecref, vecdir)
	#Setting te geometry: origin, direction and rotation vectors
	@origin = origin
	@vecref = vecref
	@vecdir = vecdir
	
	#check that the angle is valid
	angle = @vecref.angle_between(@vecdir)
	sinus = Math.sin angle
	if (sinus == 0.0)
		UI.messagebox Traductor[STR_MsgAngleZero]
		return false
	elsif (sinus.abs > 0.95)
		UI.messagebox Traductor.s(STR_MsgAngleHigh, sprintf("%3.1f", angle * 180.0 / Math::PI))
		return false
	end
	
	#computing the key parameter of the shear transformation
	@tangent = Math.tan angle
	@plane_ortho = [@origin, @vecref]
	pt0 = @origin.offset @vecdir, 10
	pt1 = pt0.project_to_plane @plane_ortho
	@vec_offset = Geom::Vector3d.new pt1.x - @origin.x, pt1.y - @origin.y, pt1.z - @origin.z
	
	#shearing the Selection
	process_entities Sketchup.active_model.selection
	Sketchup.active_model.select_tool nil
	return true			# to tell the Protractor this is OK
end

#shear Transformation of a point (subclassed method) - This contains the Shearing algorithm!
def transform_point (pt)
	pt0 = pt.project_to_plane @plane_ortho
	d = pt.distance pt0
	d = -d if (d != 0 && @vecref.samedirection?(pt0 - pt))
	newpt = pt.offset @vec_offset, d * @tangent
end

end #class PlaneShear

#********************************************************************
# SUB CLASS for processing a NULTBox transformation.
# The class implements the method <Transform_point(pt)> based on the deformation
#indicated by the NULTBox
#********************************************************************

class NultBox
	
	
end #class NultBox

#********************************************************************
# USER INTERFACE routines.
#********************************************************************

def NULTransformation.shear_command
	model = Sketchup.active_model
	cursorpath = "cursorShear.png"
	if (model.selection.count == 0)
		UI.beep
		Traductor.messagebox STR_MsgNoSelection
	else
		s = Traductor[STR_Title]
		model.select_tool LibProtractorTool::ProtractorTool.new(PlaneShear.new, cursorpath, s)
	end
end

def NULTransformation.create_button(cmd, tooltip, iconpath)
	cmd.status_bar_text = tooltip
	icon = Sketchup.find_support_file iconpath, "Plugins"
	if (icon)
		cmd.tooltip = tooltip
		cmd.small_icon = cmd.large_icon = icon
		@tlb = UI::Toolbar.new NULT_TOOLBAR unless @tlb
		@tlb.add_item cmd
		@tlb.show
	end	

end

def NULTransformation.set_command
	menutool = UI.menu "Tools"
	menutool.add_separator
	@tlb = nil
	
	#SHEAR Command
	cmd = UI::Command.new(Traductor[STR_Title]) { NULTransformation.shear_command }
	menutool.add_item cmd
	tooltip = Traductor[STR_Title] + " --> " + Traductor[STR_Tooltip]
	NULTransformation.create_button cmd, tooltip, "iconShear.png"
	
	#CREATE NULTBOX command
	
	#showing the toolbar
	@tlb.show if @tlb
end

unless $Shear____loaded
	NULTransformation.set_command
	$Shear____loaded = true
end

end #End module NULTransformation